/* 
 * This is an interface to the mySQL database for Python.
 *   
 *   Based on prior work by David Gibson
 *   
 *
 *   Based on a prior work by Anthony Baxter
 *   Updated, fixed and extended by David Gibson working for
 *   Thawte Consulting cc, South Africa.
 *
 *   Copyright 1995 Thawte Consulting cc
 *   Portions copyright (C) 1994 Anthony Baxter.
 *
 *   Permission is hereby granted, free of charge, to any person obtaining
 *   a copy of this source file to use, copy, modify, merge, or publish it
 *   subject to the following conditions:
 *
 *   The above copyright notice and this permission notice shall be included
 *   in all copies or in any new file that contains a substantial portion of
 *   this file.
 *
 *   THE  AUTHOR  MAKES  NO  REPRESENTATIONS ABOUT  THE  SUITABILITY  OF
 *   THE  SOFTWARE FOR  ANY  PURPOSE.  IT IS  PROVIDED  "AS IS"  WITHOUT
 *   EXPRESS OR  IMPLIED WARRANTY.  THE AUTHOR DISCLAIMS  ALL WARRANTIES
 *   WITH  REGARD TO  THIS  SOFTWARE, INCLUDING  ALL IMPLIED  WARRANTIES
 *   OF   MERCHANTABILITY,  FITNESS   FOR  A   PARTICULAR  PURPOSE   AND
 *   NON-INFRINGEMENT  OF THIRD  PARTY  RIGHTS. IN  NO  EVENT SHALL  THE
 *   AUTHOR  BE LIABLE  TO  YOU  OR ANY  OTHER  PARTY  FOR ANY  SPECIAL,
 *   INDIRECT,  OR  CONSEQUENTIAL  DAMAGES  OR  ANY  DAMAGES  WHATSOEVER
 *   WHETHER IN AN  ACTION OF CONTRACT, NEGLIGENCE,  STRICT LIABILITY OR
 *   ANY OTHER  ACTION ARISING OUT OF  OR IN CONNECTION WITH  THE USE OR
 *   PERFORMANCE OF THIS SOFTWARE.
 *
 * $Id: mysqlmodule.c,v 1.5 1995/12/26 12:29:25 davidg Exp davidg $

******************************************************

Ported to mySQL Skinner (joe@earthlight.co.nz) Janurary 1997

-- STATUS : BETA

--VERSION 0.1.0 
- converted source
- added support for some of the types not included in mSQL
- added support for return of auto_increment values 
	an auto_increment value is returned as a result from an insert

VERSION 0.1.2 (1997-04-02)
- added support for varchar
- added support for username and passwords
 
-- TODO 
- support for blobs
- support timestamps
- support all unsupported mysql types
- creating and destroying databases

Modified by David Gibson December 1995

- listdbs and listtables now return a list of strings
- new Python naming conventions introduced
- queries now return data types in native Python format (String,Float,Int)
- solved spurious 'function requires at least one argument' error: old
	getargs would not handle optional arguments.  ParseTuple is being used now.
	(so method table has got 1's in now)
	(old Parse routine still needed for subscript handling)
- mysql_free_result now called after query!
- assignment to subscript trapped correctly.  Ditto len()
- added DbType to the module dictionary
- mySQL.error object introduced

 */

/* #define DEBUG_MYSQLMOD */

#define INCLUDE_DATE_TIME_SUPPORT 1
/* #undef INCLUDE_DATE_TIME_SUPPORT */

#include "allobjects.h"
#include "modsupport.h"
#include <sys/types.h>
#include <sys/stat.h>

#include "mysql.h"


char   PROGNAME[] = "python";
static PyObject *pythonify_res();
static PyObject *pythonify_single_res();
static PyObject *pythonify_lf_res();

typedef struct {
    PyObject_HEAD
    MYSQL handle;
    int valid;
} mysqlobject;

staticforward PyTypeObject MsqlType;
static PyObject * mySQLError;

#define is_mysqlobject(v) ((v)->ob_type == &MsqlType)

static PyObject * 
mysqlmod_connect(self,args)
PyObject *self,*args;
{
    char *dbname = NULL;
    char *dbuser = NULL;
    char *dbpass = NULL;
    mysqlobject *n;
    MYSQL newhandle;

    if(!PyArg_ParseTuple(args,"|sss",&dbname, &dbuser, &dbpass)) {
        PyErr_SetString(mySQLError,"connect has three optional arg, a database host name, username and password");
        return NULL;
    }

    mysql_connect(&newhandle, dbname, dbuser, dbpass);

    if(newhandle.net.error) { /*JWS changed -1 to NULL */ 
        PyErr_SetString(mySQLError,mysql_error(&newhandle)); 
        return NULL;
    } else {
        n=PyObject_NEW(mysqlobject,&MsqlType);
        if(!n)
            return NULL;
        n->valid=1;
        n->handle = newhandle;
        return((PyObject *)n);
    }
}
static PyObject *
mysqlobj_selectdb(self,args)
mysqlobject *self;
PyObject *args;
{
    char *dbname;
    if(!PyArg_ParseTuple(args,"s",&dbname)) {
        PyErr_SetString(mySQLError,"selectdb has one arg, a database name"); 
        return NULL;
    }
    if(mysql_select_db(&(self->handle),dbname)==-1) {
        PyErr_SetString(mySQLError,mysql_error(&(self->handle)));
        return NULL;
    } 
    Py_INCREF(Py_None);
    return(Py_None);
}   

static PyObject * 
mysqlobj_listdbs(self,args)
mysqlobject *self;
PyObject *args;
{
    MYSQL_RES *res;
    PyObject *resobj;

    if(!PyArg_ParseTuple(args,"")) {
        PyErr_SetString(mySQLError,"listdbs takes no args");
        return NULL;
    }
    if((res=mysql_list_dbs(&(self->handle), NULL))==NULL) { /* JWS added NULL */
        Py_INCREF(Py_None);
        return(Py_None);
    }
    resobj=pythonify_single_res(res);
    return(resobj);
}

static PyObject * 
mysqlobj_listtables(self,args)
mysqlobject *self;
PyObject *args;
{
    MYSQL_RES *res;
    PyObject *resobj;

    if(!PyArg_ParseTuple(args,"")) {
        PyErr_SetString(mySQLError,"listtables takes no args");
        return NULL;
    }
    if((res=mysql_list_tables(&(self->handle), NULL))==NULL) {
        Py_INCREF(Py_None);
        return(Py_None);
    }
    resobj=pythonify_single_res(res);
    return(resobj);
}

static PyObject *
mysqlobj_listfields(self,args)
mysqlobject *self;
PyObject *args;
{
    char *tname;
    MYSQL_RES *res;
    PyObject *resobj;

    if(!PyArg_ParseTuple(args,"s",&tname)) {
        PyErr_SetString(mySQLError,"listfields takes one arg, a table name");
        return NULL;
    }
    if((res=mysql_list_fields(&(self->handle),tname, NULL))==NULL) {
        Py_INCREF(Py_None);
        return(Py_None);
    }
    resobj=pythonify_lf_res(res);
    return(resobj);
}

static PyObject * 
mysqlobj_query_helper(self,query)
mysqlobject *self;
char * query;
{
    MYSQL_RES *res;
    PyObject *resobj;

    if(mysql_query(&(self->handle),query)==-1) {
        PyErr_SetString(mySQLError,mysql_error(&(self->handle))); 
        return NULL;
    }
    res=mysql_store_result(&(self->handle));
    if(!res) { 

	/* new code to handle returning of auto_increment keys */
		if (mysql_error(&(self->handle))[0] == 0 && 
			mysql_num_fields(&(self->handle)) == 0 &&
			mysql_insert_id(&(self->handle)) != 0) 
		{
			ulong used_id;

		  	used_id = mysql_insert_id(&(self->handle));
			
			return(PyInt_FromLong(used_id));
		}
		else 
		{
			Py_INCREF(Py_None);
			return(Py_None);
		}
    }
    resobj=pythonify_res(res);
	mysql_free_result(res);
    return(resobj);
}

static PyObject * 
mysqlobj_query(self,args)
mysqlobject *self;
PyObject *args;
{
    char *query;

    if(!PyArg_ParseTuple(args,"s",&query)) {
        PyErr_SetString(mySQLError,"query has one arg, a query string");
        return NULL;
    }
	return mysqlobj_query_helper(self, query);
}

static PyObject *
mysqlobj_create(self, args)
mysqlobject *self;
PyObject *args;
{
   char *dbname;

    if(!PyArg_ParseTuple(args,"s",&dbname)) {
        PyErr_SetString(mySQLError,"createdb has one arg, a database name");
        return NULL;
    }

    if(mysql_create_db(&(self->handle),dbname)==-1) {
        PyErr_SetString(mySQLError,mysql_error(&(self->handle)));
        return NULL;
    }
    Py_INCREF(Py_None);
    return(Py_None);
}

static PyObject *
mysqlobj_drop(self, args)
mysqlobject *self;
PyObject *args;
{
   char *dbname;

    if(!PyArg_ParseTuple(args,"s",&dbname)) {
        PyErr_SetString(mySQLError,"dropdb has one arg, a database name");
        return NULL;
    }

    if(mysql_drop_db(&(self->handle),dbname)==-1) {
        PyErr_SetString(mySQLError,mysql_error(&(self->handle)));
        return NULL;
    }
    Py_INCREF(Py_None);
    return(Py_None);
}
			
static PyObject *
mysqlobj_reload(self)
mysqlobject *self;
{
    if(mysql_reload(&(self->handle))==-1) {
        PyErr_SetString(mySQLError,mysql_error(&(self->handle)));
        return NULL;
    }
    Py_INCREF(Py_None);
    return(Py_None);
}

static PyObject *
mysqlobj_shutdown(self)
mysqlobject *self;
{
    if(mysql_shutdown(&(self->handle))==-1) {
        PyErr_SetString(mySQLError,mysql_error(&(self->handle)));
        return NULL;
    }
    Py_INCREF(Py_None);
    return(Py_None);
}


/*
 * Take an mySQL MYSQL_RES, turn it into a list of tuples.
 */
static PyObject *
pythonify_res(res)
MYSQL_RES *res;
{
    PyObject *reslist,*rowtuple, *fieldobj;
    MYSQL_ROW thisrow;
    MYSQL_FIELD *tf;
    int i,n;

#ifdef DEBUG_MYSQLMOD
    printf("data ready, %d rows of %d fields\n",mysql_num_rows(res),mysql_num_fields(res));
#endif
    reslist=PyList_New(0);
    n=mysql_num_fields(res);

    while(thisrow=mysql_fetch_row(res)) {
	uint *lengths;

	lengths = mysql_fetch_lengths(res);
        rowtuple=PyTuple_New(n);
	mysql_field_seek(res, 0); /* JWS use the mysql macro */
        for (i=0;i<n;i++) {

            tf=mysql_fetch_field(res);
            if (thisrow[i]) 


/*

mysql types

 FIELD_TYPE_DECIMAL, FIELD_TYPE_CHAR,
 FIELD_TYPE_SHORT,  FIELD_TYPE_LONG,
 FIELD_TYPE_FLOAT,  FIELD_TYPE_DOUBLE,
 FIELD_TYPE_NULL,   FIELD_TYPE_TIME,
 FIELD_TYPE_LONGLONG,FIELD_TYPE_INT24,
 FIELD_TYPE_TINY_BLOB=249,
 FIELD_TYPE_MEDIUM_BLOB=250,
 FIELD_TYPE_LONG_BLOB=251,
 FIELD_TYPE_BLOB=252,
 FIELD_TYPE_VAR_STRING=253,
 FIELD_TYPE_STRING=254



*/
	
                switch (tf->type) {
                    case FIELD_TYPE_SHORT: 
                    case FIELD_TYPE_LONG:
                        fieldobj = PyInt_FromLong(atol(thisrow[i]));
                        break;
                    case FIELD_TYPE_CHAR: 
		    case FIELD_TYPE_STRING:
 		    case FIELD_TYPE_VAR_STRING:
#ifdef INCLUDE_DATE_TIME_SUPPORT 
                    case FIELD_TYPE_DATE:
	            case FIELD_TYPE_TIME:
                    case FIELD_TYPE_DATETIME:
                    case  FIELD_TYPE_TIMESTAMP:
#endif
                        fieldobj = PyString_FromString(thisrow[i]);
                        break;
 		    case FIELD_TYPE_TINY_BLOB:
		    case FIELD_TYPE_MEDIUM_BLOB:
	            case FIELD_TYPE_LONG_BLOB:
	            case FIELD_TYPE_BLOB:
                        fieldobj = PyString_FromStringAndSize(thisrow[i], 
				lengths[i]);
                        break;
                    case FIELD_TYPE_DOUBLE: 
                        fieldobj = PyFloat_FromDouble(atof(thisrow[i]));
                        break;
                    default: 
                        fieldobj = NULL;
                        break;
                }
            else {
                Py_INCREF(Py_None);
                fieldobj = Py_None;
            }

            if (fieldobj == NULL) {
                Py_DECREF(rowtuple);
                Py_DECREF(reslist);
                PyErr_SetString(mySQLError,"mySQL database returned bad value"); 
                return NULL;
            }

            PyTuple_SetItem(rowtuple,i,fieldobj);
        }
        PyList_Append(reslist,rowtuple);
        Py_DECREF(rowtuple);
    }
    return(reslist);
}

/*
 * Take an mySQL MYSQL_RES, turn it into a list of strings.
 */
static PyObject *
pythonify_single_res(res)
MYSQL_RES *res;
{
    PyObject *reslist,*rowtuple, *str;
    MYSQL_ROW thisrow;
    int i,n;

#ifdef DEBUG_MYSQLMOD
    printf("data ready, %d rows of %d fields\n",mysql_num_rows(res),mysql_num_fields(res));
#endif
    reslist=PyList_New(0);
    n=mysql_num_fields(res);
     if (n != 1) {
        PyErr_SetString(mySQLError,"expected mySQL to return singletons"); 
        return NULL;
    }

    while(thisrow=mysql_fetch_row(res)) {
          str=PyString_FromString(thisrow[0]);
          PyList_Append(reslist,str);
    }
    return(reslist);
}

/*
 * Take an mySQL MYSQL_RES, return a list of tuples of the FetchField data.
 */
static PyObject *
pythonify_lf_res(res)
MYSQL_RES *res;
{
    PyObject *reslist, *thistuple, *str;
    int i,n;
    char *type,flags[14];
    MYSQL_FIELD *tf;
    
#ifdef DEBUG_MYSQLMOD
    printf("data ready, %d fields\n", mysql_num_fields(res));
#endif
    reslist=PyList_New(0);
    n=mysql_num_fields(res);
    for(i=0;i<n;i++) {

        tf=mysql_fetch_field(res);
#ifdef DEBUG_MYSQLMOD
    printf("value of tf->type is %d \n",tf->type);
#endif
        switch(tf->type) {
            case FIELD_TYPE_SHORT: type="short"; break;
            case FIELD_TYPE_LONG: type="long"; break;
            case FIELD_TYPE_CHAR: type="char"; break;
            case FIELD_TYPE_DOUBLE: type="double"; break;
            case FIELD_TYPE_TINY_BLOB: type="tiny blob"; break;
            case FIELD_TYPE_MEDIUM_BLOB: type="medium blob"; break;
            case FIELD_TYPE_LONG_BLOB: type="long blob"; break;
 	    case FIELD_TYPE_BLOB: type="blob"; break;
#ifdef INCLUDE_DATE_TIME_SUPPORT 
            case FIELD_TYPE_DATE: type="date"; break;
	    case FIELD_TYPE_TIME: type="time"; break;
            case FIELD_TYPE_DATETIME: type="datetime"; break;
            case FIELD_TYPE_TIMESTAMP: type="timestamp"; break;
#endif
	    case FIELD_TYPE_DECIMAL:
            case FIELD_TYPE_FLOAT:
            case FIELD_TYPE_NULL:
            case FIELD_TYPE_LONGLONG:
            case FIELD_TYPE_INT24: type="unhandled"; break;
            case FIELD_TYPE_VAR_STRING: type="varchar"; break; 
            case FIELD_TYPE_STRING: type="string"; break;
		
            default: type="????"; break;
        }
        if (IS_PRI_KEY(tf->flags)) 
            strcpy(flags,"pri");
        else 
            flags[0]=0;
        if (IS_NOT_NULL(tf->flags)) 
            if(flags[0])
                strcat(flags," notnull");
            else
                strcpy(flags,"notnull");
        else 
            flags[0]=0;

        thistuple=Py_BuildValue("(sssis)",tf->name,tf->table,type,tf->length,
                                                                flags);
        PyList_Append(reslist,thistuple);
        Py_DECREF(thistuple);
    }
    return(reslist);
}

static struct PyMethodDef mysqlobj_methods[] = {
    {"selectdb",(PyCFunction)mysqlobj_selectdb, 1},
    {"query",(PyCFunction)mysqlobj_query, 1},
    {"listdbs",(PyCFunction)mysqlobj_listdbs, 1},
    {"listtables",(PyCFunction)mysqlobj_listtables, 1},
    {"listfields",(PyCFunction)mysqlobj_listfields, 1},
    {"create", (PyCFunction)mysqlobj_create, 1},
    {"drop", (PyCFunction)mysqlobj_drop, 1},
    {"reload", (PyCFunction)mysqlobj_reload, 1},
    {"shutdown", (PyCFunction)mysqlobj_shutdown, 1},
    {NULL, NULL}           /* sentinel */
};

static PyObject *
mysqlobj_getattr(ms, name)
        mysqlobject *ms;
        char *name;
{
        return Py_FindMethod(mysqlobj_methods, (PyObject *)ms, name);
}
static void     
mysqlobj_dealloc(m)
        register mysqlobject *m;
{
        if ( m->valid )
          mysql_close(&(m->handle));
        PyMem_DEL(m);
}       

int
mysqlobj_print(sp, fp, flags)
mysqlobject *sp;
FILE *fp;
int flags;
{
    fprintf(fp,"mySQL handle");
    return 0;
}

static int
mysqlobj_len(self,subs)
PyObject *self, *subs;
{
    PyErr_SetString(mySQLError,"can't take length of an mySQL handle");
    return -1;
}
static PyObject *
mysqlobj_subscript(self,subs)
PyObject *self, *subs;
{
	char * query;

    if(!PyArg_Parse(subs,"s",&query)) {
        PyErr_SetString(mySQLError,"subscript expects a query string");
        return NULL;
    }
    return mysqlobj_query_helper(self,query);
}

static int
mysqlobj_ass_sub(self,subs,val)
PyObject *self, *subs,*val;
{
    PyErr_SetString(mySQLError,"can't assign to an mySQL handle (use query insert)");
    return -1;     /* -1 is error code in interpreter main loop! (ceval.c) */
}

static PyMappingMethods mysql_as_mapping = {
        (inquiry)mysqlobj_len,            /*length*/
        (binaryfunc)mysqlobj_subscript,      /*subscript*/
        (objobjargproc)mysqlobj_ass_sub,     /*assign subscript*/
};              


static PyTypeObject MsqlType = {
        PyObject_HEAD_INIT(&PyType_Type)
        0,
        "mysqlobject",
        sizeof(mysqlobject),
        0,
        (destructor)mysqlobj_dealloc,  /*tp_dealloc*/
        0,                            /*tp_print*/
        (getattrfunc)mysqlobj_getattr, /*tp_getattr*/
        0,                            /*tp_setattr*/
        0,                            /*tp_compare*/
        0,                            /*tp_repr*/
        0,                            /*tp_as_number*/
        0,                            /*tp_as_sequence*/
        &mysql_as_mapping,             /*tp_as_mapping*/
};      


static struct PyMethodDef mysql_methods[] = {
    {"connect",mysqlmod_connect, 1},
    {NULL,NULL}
};


void
initmySQL()
{
    PyObject *module, *dict;
    char *err;

    module = Py_InitModule("mySQL", mysql_methods);
    dict = PyModule_GetDict(module);

	if (PyDict_SetItemString(dict, "DbType", (PyObject*)&MsqlType) != 0)
		Py_FatalError("Cannot add to mySQL dictionary");

	mySQLError = PyString_FromString("mySQL.error");
	if (PyDict_SetItemString(dict, "error", mySQLError) != 0)
		Py_FatalError("Cannot add to mySQL dictionary");
}
